"""
    pygments.lexers.text
    ~~~~~~~~~~~~~~~~~~~~

    Lexers for non-source code file types.

    :copyright: Copyright 2006-2021 by the Pygments team, see AUTHORS.
    :license: BSD, see LICENSE for details.
"""

from pygments.lexers.configs import ApacheConfLexer
from pygments.lexers.configs import IniLexer
from pygments.lexers.configs import LighttpdConfLexer
from pygments.lexers.configs import NginxConfLexer
from pygments.lexers.configs import PropertiesLexer
from pygments.lexers.configs import RegeditLexer
from pygments.lexers.configs import SquidConfLexer
from pygments.lexers.console import PyPyLogLexer
from pygments.lexers.data import YamlLexer
from pygments.lexers.diff import DarcsPatchLexer
from pygments.lexers.diff import DiffLexer
from pygments.lexers.haxe import HxmlLexer
from pygments.lexers.installers import DebianControlLexer
from pygments.lexers.installers import SourcesListLexer
from pygments.lexers.make import BaseMakefileLexer
from pygments.lexers.make import CMakeLexer
from pygments.lexers.make import MakefileLexer
from pygments.lexers.markup import BBCodeLexer
from pygments.lexers.markup import GroffLexer
from pygments.lexers.markup import MoinWikiLexer
from pygments.lexers.markup import RstLexer
from pygments.lexers.markup import TexLexer
from pygments.lexers.sgf import SmartGameFormatLexer
from pygments.lexers.textedit import VimLexer
from pygments.lexers.textfmts import GettextLexer
from pygments.lexers.textfmts import HttpLexer
from pygments.lexers.textfmts import IrcLogsLexer

__all__ = []
