"""
    pygments.lexers.math
    ~~~~~~~~~~~~~~~~~~~~

    Just export lexers that were contained in this module.

    :copyright: Copyright 2006-2021 by the Pygments team, see AUTHORS.
    :license: BSD, see LICENSE for details.
"""

from pygments.lexers.algebra import MuPADLexer
from pygments.lexers.idl import IDLLexer
from pygments.lexers.julia import JuliaConsoleLexer
from pygments.lexers.julia import JuliaLexer
from pygments.lexers.matlab import MatlabLexer
from pygments.lexers.matlab import MatlabSessionLexer
from pygments.lexers.matlab import OctaveLexer
from pygments.lexers.matlab import ScilabLexer
from pygments.lexers.modeling import BugsLexer
from pygments.lexers.modeling import JagsLexer
from pygments.lexers.modeling import StanLexer
from pygments.lexers.python import NumPyLexer
from pygments.lexers.r import RConsoleLexer
from pygments.lexers.r import RdLexer
from pygments.lexers.r import SLexer

__all__ = []
