"""
    pygments.lexers.compiled
    ~~~~~~~~~~~~~~~~~~~~~~~~

    Just export lexer classes previously contained in this module.

    :copyright: Copyright 2006-2021 by the Pygments team, see AUTHORS.
    :license: BSD, see LICENSE for details.
"""

from pygments.lexers.basic import BlitzBasicLexer
from pygments.lexers.basic import BlitzMaxLexer
from pygments.lexers.basic import MonkeyLexer
from pygments.lexers.business import CobolFreeformatLexer
from pygments.lexers.business import CobolLexer
from pygments.lexers.c_cpp import CLexer
from pygments.lexers.c_cpp import CppLexer
from pygments.lexers.c_like import CudaLexer
from pygments.lexers.c_like import ECLexer
from pygments.lexers.c_like import ValaLexer
from pygments.lexers.crystal import CrystalLexer
from pygments.lexers.d import DLexer
from pygments.lexers.dylan import DylanConsoleLexer
from pygments.lexers.dylan import DylanLexer
from pygments.lexers.dylan import DylanLidLexer
from pygments.lexers.felix import FelixLexer
from pygments.lexers.fortran import FortranLexer
from pygments.lexers.go import GoLexer
from pygments.lexers.graphics import GLShaderLexer
from pygments.lexers.jvm import JavaLexer
from pygments.lexers.jvm import ScalaLexer
from pygments.lexers.ml import OcamlLexer
from pygments.lexers.nimrod import NimrodLexer
from pygments.lexers.objective import LogosLexer
from pygments.lexers.objective import ObjectiveCLexer
from pygments.lexers.objective import ObjectiveCppLexer
from pygments.lexers.ooc import OocLexer
from pygments.lexers.pascal import AdaLexer
from pygments.lexers.pascal import DelphiLexer
from pygments.lexers.pascal import Modula2Lexer
from pygments.lexers.prolog import PrologLexer
from pygments.lexers.python import CythonLexer
from pygments.lexers.rust import RustLexer

__all__ = []
