"""
    pygments.lexers.functional
    ~~~~~~~~~~~~~~~~~~~~~~~~~~

    Just export lexer classes previously contained in this module.

    :copyright: Copyright 2006-2021 by the Pygments team, see AUTHORS.
    :license: BSD, see LICENSE for details.
"""

from pygments.lexers.erlang import ElixirConsoleLexer
from pygments.lexers.erlang import ElixirLexer
from pygments.lexers.erlang import ErlangLexer
from pygments.lexers.erlang import ErlangShellLexer
from pygments.lexers.haskell import HaskellLexer
from pygments.lexers.haskell import KokaLexer
from pygments.lexers.haskell import LiterateHaskellLexer
from pygments.lexers.lisp import CommonLispLexer
from pygments.lexers.lisp import NewLispLexer
from pygments.lexers.lisp import RacketLexer
from pygments.lexers.lisp import SchemeLexer
from pygments.lexers.lisp import ShenLexer
from pygments.lexers.ml import OcamlLexer
from pygments.lexers.ml import OpaLexer
from pygments.lexers.ml import SMLLexer
from pygments.lexers.theorem import CoqLexer

__all__ = []
