"""
    pygments.lexers.agile
    ~~~~~~~~~~~~~~~~~~~~~

    Just export lexer classes previously contained in this module.

    :copyright: Copyright 2006-2021 by the Pygments team, see AUTHORS.
    :license: BSD, see LICENSE for details.
"""

from pygments.lexers.d import CrocLexer
from pygments.lexers.d import MiniDLexer
from pygments.lexers.factor import FactorLexer
from pygments.lexers.iolang import IoLexer
from pygments.lexers.jvm import ClojureLexer
from pygments.lexers.jvm import IokeLexer
from pygments.lexers.lisp import SchemeLexer
from pygments.lexers.perl import Perl6Lexer
from pygments.lexers.perl import PerlLexer
from pygments.lexers.python import DgLexer
from pygments.lexers.python import Python3Lexer
from pygments.lexers.python import Python3TracebackLexer
from pygments.lexers.python import PythonConsoleLexer
from pygments.lexers.python import PythonLexer
from pygments.lexers.python import PythonTracebackLexer
from pygments.lexers.ruby import FancyLexer
from pygments.lexers.ruby import RubyConsoleLexer
from pygments.lexers.ruby import RubyLexer
from pygments.lexers.scripting import LuaLexer
from pygments.lexers.scripting import MoonScriptLexer
from pygments.lexers.tcl import TclLexer

__all__ = []
